@@ -9,22 +9,21 @@
 #include "lzfu.h"
 #include "msg.h"
 
-#define OUTPUT_TEMPLATE "%s"
+#define OUTPUT_TEMPLATE "%s.%s"
 #define OUTPUT_KMAIL_DIR_TEMPLATE ".%s.directory"
-#define KMAIL_INDEX ".%s.index"
+#define KMAIL_INDEX "../.%s.index"
 #define SEP_MAIL_FILE_TEMPLATE "%i%s"
 
 // max size of the c_time char*. It will store the date of the email
 #define C_TIME_SIZE 500
 
 struct file_ll {
-    char *name;
+    char *name[PST_TYPE_MAX];
     char *dname;
-    FILE * output;
+    FILE * output[PST_TYPE_MAX];
     int32_t stored_count;
     int32_t item_count;
     int32_t skip_count;
-    int32_t type;
 };
 
 int       grim_reaper();
@@ -34,13 +33,13 @@ void      write_email_body(FILE *f, char *body);
 void      removeCR(char *c);
 void      usage();
 void      version();
-char*     mk_kmail_dir(char* fname);
+void      mk_kmail_dir(char* fname);
 int       close_kmail_dir();
-char*     mk_recurse_dir(char* dir, int32_t folder_type);
+void      mk_recurse_dir(char* dir);
 int       close_recurse_dir();
-char*     mk_separate_dir(char *dir);
+void      mk_separate_dir(char *dir);
 int       close_separate_dir();
-void      mk_separate_file(struct file_ll *f, char *extension, int openit);
+void      mk_separate_file(struct file_ll *f, int32_t t, char *extension, int openit);
 void      close_separate_file(struct file_ll *f);
 char*     my_stristr(char *haystack, char *needle);
 void      check_filename(char *fname);
@@ -70,7 +69,6 @@ void      close_enter_dir(struct file_ll *f);
 
 const char*  prog_name;
 char*  output_dir = ".";
-char*  kmail_chdir = NULL;
 
 // Normal mode just creates mbox format files in the current directory. Each file is named
 // the same as the folder's name that it represents
@@ -180,6 +178,9 @@ int grim_reaper(int waitall)
                     //printf("Process %d terminated with signal %d\n", child, sig);
                     //fflush(stdout);
                 }
+                if (status != 0) {
+                    exit(status);
+                }
                 // this has terminated, remove it from the list
                 for (j=i; j<active_children-1; j++) {
                     child_processes[j] = child_processes[j+1];
@@ -203,7 +204,8 @@ pid_t try_fork(char *folder)
 #ifdef HAVE_FORK
 #ifdef HAVE_SEMAPHORE_H
     int available = grim_reaper(0);
-    if (available) {
+    // If children have called sem_post but not exited yet, we could have available > 0 but active_children == max_children
+    if (available && active_children < max_children) {
         sem_wait(global_children);
         pid_t child = fork();
         if (child < 0) {
@@ -240,7 +242,6 @@ void process(pst_item *outeritem, pst_desc_tree *d_ptr)
     pst_item *item = NULL;
 
     DEBUG_ENT("process");
-    memset(&ff, 0, sizeof(ff));
     create_enter_dir(&ff, outeritem);
 
     for (; d_ptr; d_ptr = d_ptr->next) {
@@ -304,25 +305,18 @@ void process(pst_item *outeritem, pst_desc_tree *d_ptr)
                 DEBUG_INFO(("skipping contact: not in output type list\n"));
             }
             else {
-                if (!ff.type) ff.type = item->type;
-                if ((ff.type != PST_TYPE_CONTACT) && (mode != MODE_SEPARATE)) {
-                    ff.skip_count++;
-                    DEBUG_INFO(("I have a contact, but the folder type %"PRIi32" isn't a contacts folder. Skipping it\n", ff.type));
+                ff.item_count++;
+                if (mode == MODE_SEPARATE) mk_separate_file(&ff, PST_TYPE_CONTACT, (mode_EX) ? ".vcf" : "", 1);
+                if (contact_mode == CMODE_VCARD) {
+                    pst_convert_utf8_null(item, &item->comment);
+                    write_vcard(ff.output[PST_TYPE_CONTACT], item, item->contact, item->comment.str);
                 }
                 else {
-                    ff.item_count++;
-                    if (mode == MODE_SEPARATE) mk_separate_file(&ff, (mode_EX) ? ".vcf" : "", 1);
-                    if (contact_mode == CMODE_VCARD) {
-                        pst_convert_utf8_null(item, &item->comment);
-                        write_vcard(ff.output, item, item->contact, item->comment.str);
-                    }
-                    else {
-                        pst_convert_utf8(item, &item->contact->fullname);
-                        pst_convert_utf8(item, &item->contact->address1);
-                        fprintf(ff.output, "%s <%s>\n", item->contact->fullname.str, item->contact->address1.str);
-                    }
-                    if (mode == MODE_SEPARATE) close_separate_file(&ff);
+                    pst_convert_utf8(item, &item->contact->fullname);
+                    pst_convert_utf8(item, &item->contact->address1);
+                    fprintf(ff.output[PST_TYPE_CONTACT], "%s <%s>\n", item->contact->fullname.str, item->contact->address1.str);
                 }
+                if (mode == MODE_SEPARATE) close_separate_file(&ff);
             }
 
         } else if (item->email && ((item->type == PST_TYPE_NOTE) || (item->type == PST_TYPE_SCHEDULE) || (item->type == PST_TYPE_REPORT))) {
@@ -332,46 +326,39 @@ void process(pst_item *outeritem, pst_desc_tree *d_ptr)
                 DEBUG_INFO(("skipping email: not in output type list\n"));
             }
             else {
-                if (!ff.type) ff.type = item->type;
-                if ((ff.type != PST_TYPE_NOTE) && (ff.type != PST_TYPE_SCHEDULE) && (ff.type != PST_TYPE_REPORT) && (mode != MODE_SEPARATE)) {
-                    ff.skip_count++;
-                    DEBUG_INFO(("I have an email type %"PRIi32", but the folder type %"PRIi32" isn't an email folder. Skipping it\n", item->type, ff.type));
-                }
-                else {
-                    char *extra_mime_headers = NULL;
-                    ff.item_count++;
-                    if (mode == MODE_SEPARATE) {
-                        // process this single email message, possibly forking
-                        pid_t parent = getpid();
-                        pid_t child = try_fork(item->file_as.str);
-                        if (child == 0) {
-                            // we are the child process, or the original parent if no children were available
-                            pid_t me = getpid();
-                            mk_separate_file(&ff, (mode_EX) ? ".eml" : "", 1);
-                            write_normal_email(ff.output, ff.name, item, mode, mode_MH, &pstfile, save_rtf_body, 0, &extra_mime_headers);
-                            close_separate_file(&ff);
-                            if (mode_MSG) {
-                                mk_separate_file(&ff, ".msg", 0);
-                                write_msg_email(ff.name, item, &pstfile);
-                            }
+                char *extra_mime_headers = NULL;
+                ff.item_count++;
+                if (mode == MODE_SEPARATE) {
+                    // process this single email message, possibly forking
+                    pid_t parent = getpid();
+                    pid_t child = try_fork(item->file_as.str);
+                    if (child == 0) {
+                        // we are the child process, or the original parent if no children were available
+                        pid_t me = getpid();
+                        mk_separate_file(&ff, PST_TYPE_NOTE, (mode_EX) ? ".eml" : "", 1);
+                        write_normal_email(ff.output[PST_TYPE_NOTE], ff.name[PST_TYPE_NOTE], item, mode, mode_MH, &pstfile, save_rtf_body, PST_TYPE_NOTE, &extra_mime_headers);
+                        close_separate_file(&ff);
+                        if (mode_MSG) {
+                            mk_separate_file(&ff, PST_TYPE_NOTE, ".msg", 0);
+                            write_msg_email(ff.name[PST_TYPE_NOTE], item, &pstfile);
+                        }
 #ifdef HAVE_FORK
 #ifdef HAVE_SEMAPHORE_H
-                            if (me != parent) {
-                                // we really were a child, forked for the sole purpose of processing this message
-                                // free my child count slot before really exiting, since
-                                // all I am doing here is waiting for my children to exit
-                                sem_post(global_children);
-                                grim_reaper(1); // wait for all my child processes to exit - there should not be any
-                                exit(0);        // really exit
-                            }
-#endif
-#endif
+                        if (me != parent) {
+                            // we really were a child, forked for the sole purpose of processing this message
+                            // free my child count slot before really exiting, since
+                            // all I am doing here is waiting for my children to exit
+                            sem_post(global_children);
+                            grim_reaper(1); // wait for all my child processes to exit - there should not be any
+                            exit(0);        // really exit
                         }
+#endif
+#endif
                     }
-                    else {
-                        // process this single email message, cannot fork since not separate mode
-                        write_normal_email(ff.output, ff.name, item, mode, mode_MH, &pstfile, save_rtf_body, 0, &extra_mime_headers);
-                    }
+                }
+                else {
+                    // process this single email message, cannot fork since not separate mode
+                    write_normal_email(ff.output[PST_TYPE_NOTE], ff.name[PST_TYPE_NOTE], item, mode, mode_MH, &pstfile, save_rtf_body, 0, &extra_mime_headers);
                 }
             }
 
@@ -382,18 +369,11 @@ void process(pst_item *outeritem, pst_desc_tree *d_ptr)
                 DEBUG_INFO(("skipping journal entry: not in output type list\n"));
             }
             else {
-                if (!ff.type) ff.type = item->type;
-                if ((ff.type != PST_TYPE_JOURNAL) && (mode != MODE_SEPARATE)) {
-                    ff.skip_count++;
-                    DEBUG_INFO(("I have a journal entry, but the folder type %"PRIi32" isn't a journal folder. Skipping it\n", ff.type));
-                }
-                else {
-                    ff.item_count++;
-                    if (mode == MODE_SEPARATE) mk_separate_file(&ff, (mode_EX) ? ".ics" : "", 1);
-                    write_journal(ff.output, item);
-                    fprintf(ff.output, "\n");
-                    if (mode == MODE_SEPARATE) close_separate_file(&ff);
-                }
+                ff.item_count++;
+                if (mode == MODE_SEPARATE) mk_separate_file(&ff, PST_TYPE_JOURNAL, (mode_EX) ? ".ics" : "", 1);
+                write_journal(ff.output[PST_TYPE_JOURNAL], item);
+                fprintf(ff.output[PST_TYPE_JOURNAL], "\n");
+                if (mode == MODE_SEPARATE) close_separate_file(&ff);
             }
 
         } else if (item->appointment && (item->type == PST_TYPE_APPOINTMENT)) {
@@ -403,24 +383,17 @@ void process(pst_item *outeritem, pst_desc_tree *d_ptr)
                 DEBUG_INFO(("skipping appointment: not in output type list\n"));
             }
             else {
-                if (!ff.type) ff.type = item->type;
-                if ((ff.type != PST_TYPE_APPOINTMENT) && (mode != MODE_SEPARATE)) {
-                    ff.skip_count++;
-                    DEBUG_INFO(("I have an appointment, but the folder type %"PRIi32" isn't an appointment folder. Skipping it\n", ff.type));
-                }
-                else {
-                    ff.item_count++;
-                    if (mode == MODE_SEPARATE) mk_separate_file(&ff, (mode_EX) ? ".ics" : "", 1);
-                    write_schedule_part_data(ff.output, item, NULL, NULL);
-                    fprintf(ff.output, "\n");
-                    if (mode == MODE_SEPARATE) close_separate_file(&ff);
-                }
+                ff.item_count++;
+                if (mode == MODE_SEPARATE) mk_separate_file(&ff, PST_TYPE_APPOINTMENT, (mode_EX) ? ".ics" : "", 1);
+                write_schedule_part_data(ff.output[PST_TYPE_APPOINTMENT], item, NULL, NULL);
+                fprintf(ff.output[PST_TYPE_APPOINTMENT], "\n");
+                if (mode == MODE_SEPARATE) close_separate_file(&ff);
             }
 
         } else if (item->message_store) {
             // there should only be one message_store, and we have already done it
             ff.skip_count++;
-            DEBUG_WARN(("item with message store content, type %i %s folder type %i, skipping it\n", item->type, item->ascii_type, ff.type));
+            DEBUG_WARN(("item with message store content, type %i %s, skipping it\n", item->type, item->ascii_type));
 
         } else {
             ff.skip_count++;
@@ -787,18 +760,12 @@ void version() {
 }
 
 
-char *mk_kmail_dir(char *fname) {
-    //change to that directory
+void mk_kmail_dir(char *fname) {
     //make a directory based on OUTPUT_KMAIL_DIR_TEMPLATE
-    //allocate space for OUTPUT_TEMPLATE and form a char* with fname
-    //return that value
-    char *dir, *out_name, *index;
+    //change to that directory
+    char *dir, *index;
     int x;
     DEBUG_ENT("mk_kmail_dir");
-    if (kmail_chdir && chdir(kmail_chdir)) {
-        x = errno;
-        DIE(("mk_kmail_dir: Cannot change to directory %s: %s\n", kmail_chdir, strerror(x)));
-    }
     dir = pst_malloc(strlen(fname)+strlen(OUTPUT_KMAIL_DIR_TEMPLATE)+1);
     sprintf(dir, OUTPUT_KMAIL_DIR_TEMPLATE, fname);
     check_filename(dir);
@@ -808,8 +775,10 @@ char *mk_kmail_dir(char *fname) {
             DIE(("mk_kmail_dir: Cannot create directory %s: %s\n", dir, strerror(x)));
         }
     }
-    kmail_chdir = pst_realloc(kmail_chdir, strlen(dir)+1);
-    strcpy(kmail_chdir, dir);
+    if (chdir(dir)) {
+        x = errno;
+        DIE(("mk_kmail_dir: Cannot change to directory %s: %s\n", dir, strerror(x)));
+    }
     free (dir);
 
     //we should remove any existing indexes created by KMail, cause they might be different now
@@ -818,36 +787,71 @@ char *mk_kmail_dir(char *fname) {
     unlink(index);
     free(index);
 
-    out_name = pst_malloc(strlen(fname)+strlen(OUTPUT_TEMPLATE)+1);
-    sprintf(out_name, OUTPUT_TEMPLATE, fname);
     DEBUG_RET();
-    return out_name;
 }
 
 
 int close_kmail_dir() {
-    // change ..
     int x;
     DEBUG_ENT("close_kmail_dir");
-    if (kmail_chdir) { //only free kmail_chdir if not NULL. do not change directory
-        free(kmail_chdir);
-        kmail_chdir = NULL;
-    } else {
-        if (chdir("..")) {
-            x = errno;
-            DIE(("close_kmail_dir: Cannot move up dir (..): %s\n", strerror(x)));
-        }
+    if (chdir("..")) {
+        x = errno;
+        DIE(("close_kmail_dir: Cannot move up dir (..): %s\n", strerror(x)));
     }
     DEBUG_RET();
     return 0;
 }
 
 
-// this will create a directory by that name,
-// then make an mbox file inside that directory.
-char *mk_recurse_dir(char *dir, int32_t folder_type) {
+char *item_type_to_name(int32_t item_type) {
+    char *name;
+    switch (item_type) {
+        case PST_TYPE_APPOINTMENT:
+            name = "calendar";
+            break;
+        case PST_TYPE_CONTACT:
+            name = "contacts";
+            break;
+        case PST_TYPE_JOURNAL:
+            name = "journal";
+            break;
+        case PST_TYPE_STICKYNOTE:
+        case PST_TYPE_TASK:
+        case PST_TYPE_NOTE:
+        case PST_TYPE_OTHER:
+        case PST_TYPE_REPORT:
+        default:
+            name = "mbox";
+            break;
+    }
+    return name;
+}
+
+
+int32_t reduced_item_type(int32_t item_type) {
+    int32_t reduced;
+    switch (item_type) {
+        case PST_TYPE_APPOINTMENT:
+        case PST_TYPE_CONTACT:
+        case PST_TYPE_JOURNAL:
+            reduced = item_type;
+            break;
+        case PST_TYPE_STICKYNOTE:
+        case PST_TYPE_TASK:
+        case PST_TYPE_NOTE:
+        case PST_TYPE_OTHER:
+        case PST_TYPE_REPORT:
+        default:
+            reduced = PST_TYPE_NOTE;
+            break;
+    }
+    return reduced;
+}
+
+
+// this will create a directory by that name
+void mk_recurse_dir(char *dir) {
     int x;
-    char *out_name;
     DEBUG_ENT("mk_recurse_dir");
     check_filename(dir);
     if (D_MKDIR (dir)) {
@@ -856,31 +860,11 @@ char *mk_recurse_dir(char *dir, int32_t folder_type) {
             DIE(("mk_recurse_dir: Cannot create directory %s: %s\n", dir, strerror(x)));
         }
     }
-    if (chdir (dir)) {
+    if (chdir(dir)) {
         x = errno;
         DIE(("mk_recurse_dir: Cannot change to directory %s: %s\n", dir, strerror(x)));
     }
-    switch (folder_type) {
-        case PST_TYPE_APPOINTMENT:
-            out_name = strdup("calendar");
-            break;
-        case PST_TYPE_CONTACT:
-            out_name = strdup("contacts");
-            break;
-        case PST_TYPE_JOURNAL:
-            out_name = strdup("journal");
-            break;
-        case PST_TYPE_STICKYNOTE:
-        case PST_TYPE_TASK:
-        case PST_TYPE_NOTE:
-        case PST_TYPE_OTHER:
-        case PST_TYPE_REPORT:
-        default:
-            out_name = strdup("mbox");
-            break;
-    }
     DEBUG_RET();
-    return out_name;
 }
 
 
@@ -896,7 +880,7 @@ int close_recurse_dir() {
 }
 
 
-char *mk_separate_dir(char *dir) {
+void mk_separate_dir(char *dir) {
     size_t dirsize = strlen(dir) + 10;
     char dir_name[dirsize];
     int x = 0, y = 0;
@@ -949,9 +933,7 @@ char *mk_separate_dir(char *dir) {
 #endif
     }
 
-    // we don't return a filename here cause it isn't necessary.
     DEBUG_RET();
-    return NULL;
 }
 
 
@@ -967,17 +949,17 @@ int close_separate_dir() {
 }
 
 
-void mk_separate_file(struct file_ll *f, char *extension, int openit) {
+void mk_separate_file(struct file_ll *f, int32_t t, char *extension, int openit) {
     DEBUG_ENT("mk_separate_file");
-    DEBUG_INFO(("opening next file to save email\n"));
+    DEBUG_INFO(("opening next file to save email type %s\n", item_type_to_name(t)));
     if (f->item_count > 999999999) { // bigger than nine 9's
         DIE(("mk_separate_file: The number of emails in this folder has become too high to handle\n"));
     }
-    sprintf(f->name, SEP_MAIL_FILE_TEMPLATE, f->item_count, extension);
-    check_filename(f->name);
+    sprintf(f->name[t], SEP_MAIL_FILE_TEMPLATE, f->item_count, extension);
+    check_filename(f->name[t]);
     if (openit) {
-        if (!(f->output = fopen(f->name, "w"))) {
-            DIE(("mk_separate_file: Cannot open file to save email \"%s\"\n", f->name));
+        if (!(f->output[t] = fopen(f->name[t], "w"))) {
+            DIE(("mk_separate_file: Cannot open file to save email \"%s\"\n", f->name[t]));
         }
     }
     DEBUG_RET();
@@ -985,16 +967,19 @@ void mk_separate_file(struct file_ll *f, char *extension, int openit) {
 
 
 void close_separate_file(struct file_ll *f) {
+    int32_t t;
     DEBUG_ENT("close_separate_file");
-    if (f->output) {
-        struct stat st;
-        fclose(f->output);
-        stat(f->name, &st);
-        if (!st.st_size) {
-            DEBUG_WARN(("removing empty output file %s\n", f->name));
-            remove(f->name);
+    for (t=0; t<PST_TYPE_MAX; t++) {
+        if (f->output[t]) {
+            struct stat st;
+            fclose(f->output[t]);
+            stat(f->name[t], &st);
+            if (!st.st_size) {
+                DEBUG_WARN(("removing empty output file %s\n", f->name[t]));
+                remove(f->name[t]);
+            }
+            f->output[t] = NULL;
         }
-        f->output = NULL;
     }
     DEBUG_RET();
 }
@@ -1190,6 +1175,10 @@ void write_inline_attachment(FILE* f_output, pst_item_attach* attach, char *boun
     }
     fprintf(f_output, "Content-Transfer-Encoding: base64\n");
 
+    if (attach->content_id.str) {
+        fprintf(f_output, "Content-ID: <%s>\n", attach->content_id.str);
+    }
+
     if (attach->filename2.str) {
         // use the long filename, converted to proper encoding if needed.
         // it is already utf8
@@ -1211,6 +1200,17 @@ void write_inline_attachment(FILE* f_output, pst_item_attach* attach, char *boun
 }
 
 
+int  header_match(char *header, char*field) {
+    int n = strlen(field);
+    if (strncasecmp(header, field, n) == 0) return 1;   // tag:{space}
+    if ((field[n-1] == ' ') && (strncasecmp(header, field, n-1) == 0)) {
+        char *crlftab = "\r\n\t";
+        DEBUG_INFO(("Possible wrapped header = %s\n", header));
+        if (strncasecmp(header+n-1, crlftab, 3) == 0) return 1; // tag:{cr}{lf}{tab}
+    }
+    return 0;
+}
+
 int  valid_headers(char *header)
 {
     // headers are sometimes really bogus - they seem to be fragments of the
@@ -1219,23 +1219,22 @@ int  valid_headers(char *header)
     // there are surely others. the problem is - given an arbitrary character
     // string, is it a valid (or even reasonable) set of rfc822 headers?
     if (header) {
-        if ((strncasecmp(header, "X-Barracuda-URL: ", 17) == 0) ||
-            (strncasecmp(header, "X-ASG-Debug-ID: ",  16) == 0) ||
-            (strncasecmp(header, "Return-Path: ",     13) == 0) ||
-            (strncasecmp(header, "Received: ",        10) == 0) ||
-            (strncasecmp(header, "Subject: ",          9) == 0) ||
-            (strncasecmp(header, "Date: ",             6) == 0) ||
-            (strncasecmp(header, "From: ",             6) == 0) ||
-            (strncasecmp(header, "X-x: ",              5) == 0) ||
-            (strncasecmp(header, "Microsoft Mail Internet Headers", 31) == 0)) {
-            return 1;
-        }
-        else {
-            if (strlen(header) > 2) {
-                DEBUG_INFO(("Ignore bogus headers = %s\n", header));
-            }
-            return 0;
+        if (header_match(header, "Content-Type: "                 )) return 1;
+        if (header_match(header, "Date: "                         )) return 1;
+        if (header_match(header, "From: "                         )) return 1;
+        if (header_match(header, "MIME-Version: "                 )) return 1;
+        if (header_match(header, "Microsoft Mail Internet Headers")) return 1;
+        if (header_match(header, "Received: "                     )) return 1;
+        if (header_match(header, "Return-Path: "                  )) return 1;
+        if (header_match(header, "Subject: "                      )) return 1;
+        if (header_match(header, "To: "                           )) return 1;
+        if (header_match(header, "X-ASG-Debug-ID: "               )) return 1;
+        if (header_match(header, "X-Barracuda-URL: "              )) return 1;
+        if (header_match(header, "X-x: "                          )) return 1;
+        if (strlen(header) > 2) {
+            DEBUG_INFO(("Ignore bogus headers = %s\n", header));
         }
+        return 0;
     }
     else return 0;
 }
@@ -1773,7 +1772,6 @@ void write_normal_email(FILE* f_output, char f_name[], pst_item* item, int mode,
     if (item->email->encrypted_body.data) {
         pst_item_attach* attach = (pst_item_attach*)pst_malloc(sizeof(pst_item_attach));
         DEBUG_INFO(("Adding encrypted text body as attachment\n"));
-        attach = (pst_item_attach*) pst_malloc(sizeof(pst_item_attach));
         memset(attach, 0, sizeof(pst_item_attach));
         attach->next = item->attach;
         item->attach = attach;
@@ -1785,7 +1783,6 @@ void write_normal_email(FILE* f_output, char f_name[], pst_item* item, int mode,
     if (item->email->encrypted_htmlbody.data) {
         pst_item_attach* attach = (pst_item_attach*)pst_malloc(sizeof(pst_item_attach));
         DEBUG_INFO(("Adding encrypted HTML body as attachment\n"));
-        attach = (pst_item_attach*) pst_malloc(sizeof(pst_item_attach));
         memset(attach, 0, sizeof(pst_item_attach));
         attach->next = item->attach;
         item->attach = attach;
@@ -2170,15 +2167,30 @@ void write_appointment(FILE* f_output, pst_item* item)
 
 void create_enter_dir(struct file_ll* f, pst_item *item)
 {
-    pst_convert_utf8(item, &item->file_as);
-    f->type         = item->type;
+    memset(f, 0, sizeof(*f));
     f->stored_count = (item->folder) ? item->folder->item_count : 0;
+    pst_convert_utf8(item, &item->file_as);
+    f->dname = (char*) pst_malloc(strlen(item->file_as.str)+1);
+    strcpy(f->dname, item->file_as.str);
 
     DEBUG_ENT("create_enter_dir");
-    if (mode == MODE_KMAIL)
-        f->name = mk_kmail_dir(item->file_as.str);
-    else if (mode == MODE_RECURSE) {
-        f->name = mk_recurse_dir(item->file_as.str, f->type);
+    if (mode == MODE_KMAIL) {
+        int32_t t;
+        mk_kmail_dir(item->file_as.str);
+        for (t=0; t<PST_TYPE_MAX; t++) {
+            if (t == reduced_item_type(t)) {
+                f->name[t] = (char*) pst_malloc(strlen(item->file_as.str)+strlen(OUTPUT_TEMPLATE)+30);
+                sprintf(f->name[t], OUTPUT_TEMPLATE, item->file_as.str, item_type_to_name(t));
+            }
+        }
+    } else if (mode == MODE_RECURSE) {
+        int32_t t;
+        mk_recurse_dir(item->file_as.str);
+        for (t=0; t<PST_TYPE_MAX; t++) {
+            if (t == reduced_item_type(t)) {
+                f->name[t] = strdup(item_type_to_name(t));
+            }
+        }
         if (mode_thunder) {
             FILE *type_file = fopen(".type", "w");
             fprintf(type_file, "%d\n", item->type);
@@ -2186,47 +2198,58 @@ void create_enter_dir(struct file_ll* f, pst_item *item)
         }
     } else if (mode == MODE_SEPARATE) {
         // do similar stuff to recurse here.
+        int32_t t;
         mk_separate_dir(item->file_as.str);
-        f->name = (char*) pst_malloc(file_name_len);
-        memset(f->name, 0, file_name_len);
-    } else {
-        f->name = (char*) pst_malloc(strlen(item->file_as.str)+strlen(OUTPUT_TEMPLATE)+1);
-        sprintf(f->name, OUTPUT_TEMPLATE, item->file_as.str);
-    }
-
-    f->dname = (char*) pst_malloc(strlen(item->file_as.str)+1);
-    strcpy(f->dname, item->file_as.str);
-
-    if (overwrite != 1) {
-        int x = 0;
-        char *temp = (char*) pst_malloc (strlen(f->name)+10); //enough room for 10 digits
-
-        sprintf(temp, "%s", f->name);
-        check_filename(temp);
-        while ((f->output = fopen(temp, "r"))) {
-            DEBUG_INFO(("need to increase filename because one already exists with that name\n"));
-            DEBUG_INFO(("- increasing it to %s%d\n", f->name, x));
-            x++;
-            sprintf(temp, "%s%08d", f->name, x);
-            DEBUG_INFO(("- trying \"%s\"\n", f->name));
-            if (x == 99999999) {
-                DIE(("create_enter_dir: Why can I not create a folder %s? I have tried %i extensions...\n", f->name, x));
+        for (t=0; t<PST_TYPE_MAX; t++) {
+            if (t == reduced_item_type(t)) {
+                f->name[t] = (char*) pst_malloc(file_name_len);
+                memset(f->name[t], 0, file_name_len);
             }
-            fclose(f->output);
         }
-        if (x > 0) { //then the f->name should change
-            free (f->name);
-            f->name = temp;
-        } else {
-            free(temp);
+    } else {
+        // MODE_NORMAL
+        int32_t t;
+        for (t=0; t<PST_TYPE_MAX; t++) {
+            if (t == reduced_item_type(t)) {
+                f->name[t] = (char*) pst_malloc(strlen(item->file_as.str)+strlen(OUTPUT_TEMPLATE)+30);
+                sprintf(f->name[t], OUTPUT_TEMPLATE, item->file_as.str, item_type_to_name(t));
+            }
         }
     }
 
-    DEBUG_INFO(("f->name = %s\nitem->folder_name = %s\n", f->name, item->file_as.str));
     if (mode != MODE_SEPARATE) {
-        check_filename(f->name);
-        if (!(f->output = fopen(f->name, "w"))) {
-            DIE(("create_enter_dir: Could not open file \"%s\" for write\n", f->name));
+        int32_t t;
+        for (t=0; t<PST_TYPE_MAX; t++) {
+            if (f->name[t]) {
+                if (!overwrite) {
+                    int x = 0;
+                    char *temp = (char*) pst_malloc (strlen(f->name[t])+10); //enough room for 10 digits
+
+                    sprintf(temp, "%s", f->name[t]);
+                    check_filename(temp);
+                    while ((f->output[t] = fopen(temp, "r"))) {
+                        DEBUG_INFO(("need to increase filename because one already exists with that name\n"));
+                        x++;
+                        sprintf(temp, "%s%08d", f->name[t], x);
+                        DEBUG_INFO(("- bump file name and try \"%s\"\n", temp));
+                        if (x == 99999999) {
+                            DIE(("create_enter_dir: Why can I not create a folder %s? I have tried %i extensions...\n", f->name[t], x));
+                        }
+                        fclose(f->output[t]);
+                    }
+                    if (x > 0) { //then the f->name should change
+                        free (f->name[t]);
+                        f->name[t] = temp;
+                    } else {
+                        free(temp);
+                    }
+                }
+                check_filename(f->name[t]);
+                if (!(f->output[t] = fopen(f->name[t], "w"))) {
+                    DIE(("create_enter_dir: Could not open file \"%s\" for write\n", f->name[t]));
+                }
+                DEBUG_INFO(("f->name = %s\nitem->folder_name = %s\n", f->name[t], item->file_as.str));
+            }
         }
     }
     DEBUG_RET();
@@ -2235,6 +2258,7 @@ void create_enter_dir(struct file_ll* f, pst_item *item)
 
 void close_enter_dir(struct file_ll *f)
 {
+    int32_t t;
     DEBUG_INFO(("processed item count for folder %s is %i, skipped %i, total %i \n",
                 f->dname, f->item_count, f->skip_count, f->stored_count));
     if (output_mode != OUTPUT_QUIET) {
@@ -2243,18 +2267,23 @@ void close_enter_dir(struct file_ll *f)
             fflush(stdout);
         pst_debug_unlock();
     }
-    if (f->output) {
-        if (mode == MODE_SEPARATE) DEBUG_WARN(("close_enter_dir finds open separate file\n"));
-        struct stat st;
-        fclose(f->output);
-        stat(f->name, &st);
-        if (!st.st_size) {
-            DEBUG_WARN(("removing empty output file %s\n", f->name));
-            remove(f->name);
+    for (t=0; t<PST_TYPE_MAX; t++) {
+        if (f->output[t]) {
+            if (mode == MODE_SEPARATE) DEBUG_WARN(("close_enter_dir finds open separate file\n"));
+            fclose(f->output[t]);
+            f->output[t] = NULL;
+        }
+        if (f->name[t]) {
+            struct stat st;
+            stat(f->name[t], &st);
+            if (!st.st_size) {
+                DEBUG_WARN(("removing empty output file %s\n", f->name[t]));
+                remove(f->name[t]);
+            }
+            free(f->name[t]);
+            f->name[t] = NULL;
         }
-        f->output = NULL;
     }
-    free(f->name);
     free(f->dname);
 
     if (mode == MODE_KMAIL)
